import hypothesis.strategies as st
from hypothesis_cfg import ContextFreeGrammarStrategy
from parsimonious import Grammar, NodeVisitor

from aiger import common, expr

CIRC_GRAMMAR = Grammar(u'''
phi =  and / neg / vyest / AP
and = "(" _ phi _ "&" _ phi _ ")"
neg = "~" _ phi
vyest = "Z" _ phi

_ = ~r" "*
AP = ~r"[a-zA-z]" ~r"[a-zA-Z\\d]*"
EOL = "\\n"
''')


class CircVisitor(NodeVisitor):
    def generic_visit(self, _, children):
        return children

    def visit_phi(self, _, children):
        return children[0]

    def visit_AP(self, node, _):
        return expr.atom(node.text)

    def visit_and(self, _, children):
        return children[2] & children[6]

    def visit_neg(self, _, children):
        return ~children[2]

    def visit_vyest(self, _, children):
        _, _, phi = children
        aig = phi.aig >> common.delay(
            inputs=phi.aig.outputs,
            initials=[True],
            latches=[common._fresh()],
            outputs=[common._fresh()]
        )
        return expr.BoolExpr(aig)


def parse(circ_str: str):
    return CircVisitor().visit(CIRC_GRAMMAR.parse(circ_str)).aig


GRAMMAR = {
    'psi': (('(', 'psi', ' & ', 'psi', ')'), ('~ ', 'psi'), ('Z ', 'psi'),
            ('AP', )),
    'AP': (
        ('a', ),
        ('b', ),
        ('c', ),
        ('d', ),
        ('e', ),
        ('f', ),
        ('g', ),
        ('h', ),
        ('i', ),
        ('j', ),
        ('k', ),
        ('l', ),
    ),
}


def make_circuit(term):
    circ_str = ''.join(term)
    return parse(circ_str).evolve(comments=(circ_str, ))


Circuits = st.builds(make_circuit,
                     ContextFreeGrammarStrategy(
                         GRAMMAR, max_length=15, start='psi'))



import click

import aiger


@click.command()
@click.argument('path1', type=click.Path(exists=True))
@click.argument('path2', type=click.Path(exists=True))
@click.argument('dest', type=click.Path(exists=False))
def parse_and_compose(path1, path2, dest):
    aag1, aag2 = map(aiger.parser.load, (path1, path2))

    with open(dest, 'w') as f:
        f.write((aag1 >> aag2).dump())


@click.command()
@click.argument('path1', type=click.Path(exists=True))
@click.argument('path2', type=click.Path(exists=True))
@click.argument('dest', type=click.Path(exists=False))
def parse_and_parcompose(path1, path2, dest):
    aag1, aag2 = map(aiger.parser.load, (path1, path2))

    with open(dest, 'w') as f:
        f.write((aag1 | aag2).dump())



import funcy as fn
import hypothesis.strategies as st
import pytest
from hypothesis import given

from aiger import hypothesis as aigh
from aiger import common


@given(st.integers(2, 10), st.data())
def test_and(n_inputs, data):
    aag = common.and_gate([f'x{i}' for i in range(n_inputs)], "out")
    test_input = {f'x{i}': data.draw(st.booleans()) for i in range(n_inputs)}
    out, _ = aag(test_input)
    assert out['out'] == all(test_input.values())


@given(aigh.Circuits, aigh.Circuits, st.data())
def test_and2(aag1, aag2, data):
    aag3 = aag1 | aag2
    aag3 >>= common.and_gate(aag3.outputs)

    test_input = {f'{i}': data.draw(st.booleans()) for i in aag3.inputs}

    out1, _ = aag1(test_input)
    out2, _ = aag2(test_input)
    out3, _ = aag3(test_input)

    v12 = list(out1.values())[0] and list(out2.values())[0]
    v3 = list(out3.values())[0]
    assert v12 == v3


@given(st.integers(2, 10), st.data())
def test_or(n_inputs, data):
    aag = common.or_gate([f'x{i}' for i in range(n_inputs)], "out")
    test_input = {f'x{i}': data.draw(st.booleans()) for i in range(n_inputs)}
    out, _ = aag(test_input)
    assert out['out'] == any(test_input.values())


@given(aigh.Circuits, aigh.Circuits, st.data())
def test_or2(aag1, aag2, data):
    aag3 = aag1 | aag2
    aag3 >>= common.or_gate(aag3.outputs)

    test_input = {f'{i}': data.draw(st.booleans()) for i in aag3.inputs}

    out1, _ = aag1(test_input)
    out2, _ = aag2(test_input)
    out3, _ = aag3(test_input)

    v12 = list(out1.values())[0] or list(out2.values())[0]
    v3 = list(out3.values())[0]
    assert v12 == v3


@given(st.data())
def test_source(data):
    circ = common.source({'x': True, 'y': False})
    assert len(circ.inputs) == 0
    assert circ({})[0] == {'x': True, 'y': False}


@given(st.data())
def test_sink(data):
    circ = common.sink(['x', 'y'])
    test_input = {i: data.draw(st.booleans()) for i in circ.inputs}
    assert circ(test_input)[0] == dict()


@given(st.integers(1, 5), st.data())
def test_ite(n, data):
    inputs0 = [f'i0_{idx}' for idx in range(n)]
    inputs1 = [f'i1_{idx}' for idx in range(n)]
    outputs = [f'o{idx}' for idx in range(n)]

    circ = common.ite('test', inputs1, inputs0, outputs)
    assert len(circ.outputs) == n

    _inputs = {'test': data.draw(st.booleans())}
    _inputs.update({i: data.draw(st.booleans()) for i in inputs0})
    _inputs.update({i: data.draw(st.booleans()) for i in inputs1})

    res, _ = circ(_inputs)
    for i0, i1, out in zip(inputs0, inputs1, outputs):
        if _inputs['test']:
            assert res[out] == _inputs[i1]
        else:
            assert res[out] == _inputs[i0]


@given(aigh.Circuits, st.data())
def test_flipper(aag1, data):
    aag2 = aag1 >> common.bit_flipper(aag1.outputs)
    aag3 = aag2 >> common.bit_flipper(aag2.outputs)

    test_input = {f'{i}': data.draw(st.booleans()) for i in aag1.inputs}
    out1, _ = aag1(test_input)
    out2, _ = aag2(test_input)
    out3, _ = aag3(test_input)

    v1 = list(out1.values())[0]
    v2 = list(out2.values())[0]
    v3 = list(out3.values())[0]

    assert v1 == (not v2)
    assert v1 == v3


@given(aigh.Circuits, st.data())
def test_tee(aag1, data):
    # TODO
    aag2 = aag1 >> common.tee({k: ['a', 'b', 'c'] for k in aag1.outputs})

    assert len(aag2.outputs) == 3
    assert len(aag2.inputs) == len(aag1.inputs)

    test_input = {f'{i}': data.draw(st.booleans()) for i in aag1.inputs}
    out1, _ = aag1(test_input)
    out2, _ = aag2(test_input)
    v1 = list(out1.values())[0]
    for v in out2.values():
        assert v1 == v


@given(aigh.Circuits, st.data())
def test_feedback(aag1, data):
    aag2 = aag1 | (common.identity(['##test##']))
    assert len(aag2.outputs) == 1 + len(aag1.outputs)
    assert len(aag2.inputs) == 1 + len(aag1.inputs)

    aag3 = aag2.feedback(
        inputs=['##test##'],
        outputs=['##test##'],
        initials=[True],
        keep_outputs=False)

    assert aag3.outputs == aag1.outputs
    assert aag3.inputs == aag1.inputs

    test_input = {f'{i}': data.draw(st.booleans()) for i in aag3.inputs}
    assert aag1(test_input)[0] == aag3(test_input)[0]


@given(aigh.Circuits)
def test_relabel(aag1):
    # TODO
    new_inputs = {k: f'{k}#2' for k in aag1.inputs}
    assert set(aag1['i', new_inputs].inputs) == set(new_inputs.values())

    new_outputs = {k: f'{k}#2' for k in aag1.outputs}
    assert set(aag1['o', new_outputs].outputs) == set(new_outputs.values())

    with pytest.raises(AssertionError):
        aag1['z', {}]


@given(aigh.Circuits, st.data())
def test_cutlatches(aag1, data):
    aag2, lmap = aag1.cutlatches(aag1.latches)

    assert aag2.inputs >= aag1.inputs
    assert aag2.outputs >= aag1.outputs
    assert len(aag2.latches) == 0

    test_inputs = {i: data.draw(st.booleans()) for i in aag1.inputs}
    test_latch_ins = {l: data.draw(st.booleans()) for l in aag1.latches}
    test_inputs2 = fn.merge(test_inputs,
                            {lmap[k][0]: v
                             for k, v in test_latch_ins.items()})
    out_vals, latch_vals = aag1(test_inputs, latches=test_latch_ins)
    out_vals2, _ = aag2(test_inputs2)
    assert fn.project(out_vals2, aag1.outputs) == out_vals
    latch_vals2 = {k: out_vals2[v] for k, (v, _) in lmap.items()}
    assert latch_vals == latch_vals2


@given(aigh.Circuits, st.integers(min_value=1, max_value=4), st.data())
def test_unroll_simulate(aag1, horizon, data):
    # TODO
    aag2 = aag1.unroll(horizon)
    assert horizon * len(aag1.inputs) == len(aag2.inputs)
    assert horizon * len(aag1.outputs) == len(aag2.outputs)

    test_inputs = [{f'{i}': data.draw(st.booleans())
                    for i in aag1.inputs} for _ in range(horizon)]

    time = -1

    def unroll_keys(inputs):
        nonlocal time
        time += 1

        def unroll_key(key):
            return f"{key}##time_{time}"

        return fn.walk_keys(unroll_key, inputs)

    # Check the number of trues and falses match up.
    sum1 = sum(sum(x.values()) for x, _ in aag1.simulate(test_inputs))
    sum2 = sum(aag2(fn.merge(*map(unroll_keys, test_inputs)))[0].values())
    assert sum1 == sum2


def test_eval_order_smoke():
    circ = common.and_gate([f'x{i}' for i in range(16)], "out")
    assert len(common.eval_order(circ)) == 2*16 - 1



from tempfile import NamedTemporaryFile

import hypothesis.strategies as st
from hypothesis import given, settings

import aiger
from aiger import common
from aiger import hypothesis as aigh
from aiger import parser as aigp


@settings(deadline=500)
@given(aigh.Circuits, st.data())
def test_load(circ, data):
    with NamedTemporaryFile() as f:
        circ.write(f.name)
        circ2 = aigp.load(f.name)

    assert circ.inputs == circ2.inputs
    assert circ.outputs == circ2.outputs
    test_input = {f'{i}': data.draw(st.booleans()) for i in circ.inputs}
    assert circ(test_input) == circ2(test_input)


@given(aigh.Circuits, st.data())
def test_parse(circ, data):
    circ2 = aigp.parse(repr(circ))

    assert circ.inputs == circ2.inputs
    assert circ.outputs == circ2.outputs
    test_input = {f'{i}': data.draw(st.booleans()) for i in circ.inputs}
    assert circ(test_input) == circ2(test_input)


@given(aigh.Circuits, st.data())
def test_aig_to_aag(circ, data):
    circ2 = circ._to_aag()._to_aig()
    assert circ.inputs == circ2.inputs
    assert circ.outputs == circ2.outputs
    test_input = {f'{i}': data.draw(st.booleans()) for i in circ.inputs}
    assert circ(test_input) == circ2(test_input)


def test_sink_aag():
    circ = aiger.common.sink(['x', 'y'])
    assert len(circ._to_aag().inputs) != 0
    assert len(circ._to_aag().outputs) == 0
    assert len(circ._to_aag().latches) == 0
    circ2 = circ._to_aag()._to_aig()
    assert circ.inputs == circ2.inputs
    assert circ.outputs == circ2.outputs


@given(aigh.Circuits, st.data())
def test_dummylatches_aag(circ, ddata):
    circ2 = circ.evolve(
        latch2init={common._fresh(): False}
    )
    circ3 = circ._to_aag()._to_aig()
    assert circ2.inputs == circ3.inputs
    assert circ3.outputs == circ3.outputs
    assert circ3.latches == circ3.latches



import operator as op
from collections import defaultdict
from itertools import starmap
from functools import reduce
from uuid import uuid1

import funcy as fn
from toposort import toposort

from aiger import aig


def _fresh():
    return str(uuid1())


def _map_tree(inputs, f):
    queue = fn.lmap(aig.Input, inputs)
    while len(queue) > 1:
        queue = fn.lmap(f, fn.chunks(2, queue))
    return queue[0]


def _and(left_right):
    if len(left_right) == 1:
        return left_right[0]

    return aig.AndGate(*left_right)


def and_gate(inputs, output=None):
    output = f'#and_output#{hash(tuple(inputs))}' if output is None else output

    return aig.AIG(
        inputs=frozenset(inputs),
        node_map=frozenset(((output, _map_tree(inputs, f=_and)), )),
    )


def or_gate(inputs, output=None):
    output = f'#or_output#{hash(tuple(inputs))}' if output is None else output
    circ = and_gate(inputs, output)

    return bit_flipper(inputs) >> circ >> bit_flipper([output])


def _xor(left_right):
    if len(left_right) == 1:
        return left_right[0]

    return aig.AndGate(
        aig.Inverter(aig.AndGate(*left_right)),  # Both True
        aig.Inverter(aig.AndGate(*map(aig.Inverter, left_right)))  # Both False
    )


def parity_gate(inputs, output=None):
    # TODO:
    output = f'#parity#{hash(tuple(inputs))}' if output is None else output

    return aig.AIG(
        inputs=frozenset(inputs),
        node_map=frozenset(((output, _map_tree(inputs, f=_xor)), )),
    )


def identity(inputs, outputs=None):
    if outputs is None:
        outputs = inputs

    return aig.AIG(
        inputs=frozenset(inputs),
        node_map=frozenset(zip(outputs, map(aig.Input, inputs))),
    )


def empty():
    return aig.AIG()


def _inverted_input(name):
    return aig.Inverter(aig.Input(name))


def bit_flipper(inputs, outputs=None):
    if outputs is None:
        outputs = inputs
    else:
        assert len(outputs) == len(inputs)

    return aig.AIG(
        inputs=frozenset(inputs),
        node_map=frozenset(zip(outputs, map(_inverted_input, inputs))),
    )


def _const(val):
    return aig.Inverter(aig.ConstFalse()) if val else aig.ConstFalse()


def source(outputs):
    return aig.AIG(
        node_map=frozenset((k, _const(v)) for k, v in outputs.items()),
    )


def sink(inputs):
    return aig.AIG(
        inputs=frozenset(inputs),
    )


def tee(outputs=None):
    if not outputs:
        return empty()

    def tee_output(name, renames):
        return frozenset((r, aig.Input(name)) for r in renames)

    return aig.AIG(
        inputs=frozenset(outputs),
        node_map=frozenset.union(*starmap(tee_output, outputs.items())),
    )


def _ite(test: str, in1: str, in0: str, output: str = None):
    r"test -> in1 /\ ~test -> in0"
    assert len({test, in0, in1}) == 3

    true_out = bit_flipper([test]) >> or_gate([test, in1], 'true_out')
    false_out = or_gate([test, in0], 'false_out')
    return (true_out | false_out) >> and_gate(['true_out', 'false_out'],
                                              output)


def ite(test, inputs1, inputs0, outputs):
    assert len(inputs1) > 0
    assert len(inputs1) == len(inputs0) == len(outputs)
    assert len({test} | set(inputs1) | set(inputs0)) == 2 * len(inputs0) + 1

    ites = [_ite(test, *args) for args in zip(inputs1, inputs0, outputs)]
    return reduce(op.or_, ites)


def delay(inputs, initials, latches=None, outputs=None):
    if outputs is None:
        outputs = inputs

    if latches is None:
        latches = inputs

    assert len(inputs) == len(initials) == len(outputs) == len(latches)

    _inputs = map(aig.Input, inputs)
    _latches = map(aig.LatchIn, latches)
    return aig.AIG(
        inputs=frozenset(inputs),
        latch_map=frozenset(zip(latches, _inputs)),
        latch2init=frozenset(zip(latches, initials)),
        node_map=frozenset(zip(outputs, _latches)),
    )


def _dependency_graph(nodes):
    queue, deps, visited = list(nodes), defaultdict(set), set()
    while queue:
        node = queue.pop()
        if node in visited:
            continue
        else:
            visited.add(node)

        children = node.children
        queue.extend(children)
        deps[node].update(children)

    return deps


def eval_order(circ):
    return fn.lcat(toposort(_dependency_graph(circ.cones | circ.latch_cones)))



# flake8: noqa
from aiger.aig import AIG
from aiger.common import and_gate, or_gate, bit_flipper, source, sink
from aiger.common import tee, empty, identity, delay, parity_gate
from aiger.parser import parse, load
from aiger.expr import atom, ite, BoolExpr



from setuptools import find_packages, setup

DESC = 'A python library for manipulating sequential and-inverter gates.'

setup(
    name='py-aiger',
    version='3.1.0',
    description=DESC,
    url='http://github.com/mvcisback/py-aiger',
    author='Marcell Vazquez-Chanlatte',
    author_email='marcell.vc@eecs.berkeley.edu',
    license='MIT',
    entry_points={
        'console_scripts': [
            'aigseqcompose = aiger.utils:parse_and_compose',
            'aigparcompose = aiger.utils:parse_and_parcompose',
        ],
    },
    install_requires=[
        'attr',
        'bidict',
        'click',
        'funcy',
        'parsimonious',
        'toposort',
    ],
    packages=find_packages(),
)



from functools import reduce
from typing import Tuple, FrozenSet, NamedTuple, Union

import attr
import funcy as fn

from aiger import common as cmn
from aiger import parser


def timed_name(name, time):
    return f"{name}##time_{time}"


class AndGate(NamedTuple):
    left: 'Node'  # TODO: replace with Node once 3.7 lands.
    right: 'Node'

    @property
    def children(self):
        return (self.left, self.right)

    def __hash__(self):
        return id(self)


class Inverter(NamedTuple):
    input: 'Node'

    @property
    def children(self):
        return (self.input, )

    def __hash__(self):
        return id(self)


# Enables filtering for Input via lens library.
class Input(NamedTuple):
    name: str

    @property
    def children(self):
        return ()


class LatchIn(NamedTuple):
    name: str

    @property
    def children(self):
        return ()


class ConstFalse(NamedTuple):
    @property
    def children(self):
        return ()


def _is_const_true(node):
    return isinstance(node, Inverter) and isinstance(node.input, ConstFalse)


Node = Union[AndGate, ConstFalse, Inverter, Input, LatchIn]


@attr.s(frozen=True, slots=True, auto_attribs=True, repr=False)
class AIG:
    inputs: FrozenSet[str] = frozenset()
    node_map: FrozenSet[Tuple[str, Node]] = frozenset()
    latch_map: FrozenSet[Tuple[str, Node]] = frozenset()
    latch2init: FrozenSet[Tuple[str, Node]] = frozenset()
    comments: Tuple[str] = ()

    _to_aag = parser.aig2aag

    def __repr__(self):
        return repr(self._to_aag())

    def __getitem__(self, others):
        assert isinstance(others, tuple) and len(others) == 2
        kind, relabels = others
        assert kind in {'i', 'o', 'l'}

        basis = {
            'i': self.inputs, 'o': self.outputs, 'l': self.latches
        }.get(kind)
        relabels = fn.project(relabels, basis)

        if kind == 'o':
            relabels = {k: [v] for k, v in relabels.items()}
            return self >> cmn.tee(relabels)

        relabels = {v: [k] for k, v in relabels.items()}
        input_kinds = (Input,) if kind == 'i' else (LatchIn,)
        return seq_compose(cmn.tee(relabels), self, input_kinds=input_kinds)

    def evolve(self, **kwargs):
        return attr.evolve(self, **kwargs)

    @property
    def outputs(self):
        return frozenset(fn.pluck(0, self.node_map))

    @property
    def latches(self):
        return frozenset(fn.pluck(0, self.latch2init))

    @property
    def cones(self):
        return frozenset(fn.pluck(1, self.node_map))

    @property
    def latch_cones(self):
        return frozenset(fn.pluck(1, self.latch_map))

    def __rshift__(self, other):
        return seq_compose(self, other)

    def __lshift__(self, other):
        return seq_compose(other, self)

    def __or__(self, other):
        return par_compose(self, other)

    def __call__(self, inputs, latches=None):
        if latches is None:
            latches = dict()
        latchins = fn.merge(dict(self.latch2init), latches)

        def sub(node):
            if isinstance(node, ConstFalse):
                return node
            store = inputs if isinstance(node, Input) else latchins
            return Inverter(ConstFalse()) if store[node.name] else ConstFalse()

        circ = self._modify_leafs(sub)
        outputs = {n: _is_const_true(node) for n, node in circ.node_map}
        latch_outputs = {n: _is_const_true(node) for n, node in circ.latch_map}
        return outputs, latch_outputs

    def simulator(self, latches=None):
        inputs = yield
        while True:
            outputs, latches = self(inputs, latches)
            inputs = yield outputs, latches

    def simulate(self, input_seq, latches=None):
        sim = self.simulator()
        next(sim)
        return [sim.send(inputs) for inputs in input_seq]

    def cutlatches(self, latches, check_postcondition=True):
        l_map = {n: (cmn._fresh(), init) for (n, init) in self.latch2init}

        assert len(
            set(fn.pluck(0, l_map.values())) & (self.inputs | self.outputs)
        ) == 0

        def sub(node):
            if isinstance(node, LatchIn):
                return Input(l_map[node.name][0])
            return node

        circ = self._modify_leafs(sub)
        _cones = {(l_map[k][0], v) for k, v in circ.latch_map if k in latches}
        aig = self.evolve(
            node_map=circ.node_map | _cones,
            inputs=self.inputs | {n for n, _ in l_map.values()},
            latch_map={(k, v) for k, v in circ.latch_map if k not in latches},
            latch2init={(k, v) for k, v in self.latch2init if k not in latches}
        )
        return aig, l_map

    def feedback(
        self, inputs, outputs, initials=None, latches=None, keep_outputs=False
    ):
        if latches is None:
            latches = inputs

        if initials is None:
            initials = [False for _ in inputs]

        assert len(inputs) == len(initials) == len(outputs) == len(latches)
        assert len(set(inputs) & self.inputs) != 0
        assert len(set(outputs) & self.outputs) != 0

        in2latch = {iname: lname for iname, lname in zip(inputs, latches)}

        def sub(node):
            if isinstance(node, Input) and node.name in inputs:
                return LatchIn(in2latch[node.name])
            return node

        aig = self._modify_leafs(sub)

        _latch_map, node_map = fn.lsplit(
            lambda x: x[0] in outputs, aig.node_map
        )
        out2latch = {oname: lname for oname, lname in zip(outputs, latches)}
        _latch_map = {(out2latch[k], v) for k, v in _latch_map}
        l2init = frozenset((n, val) for n, val in zip(latches, initials))
        return aig.evolve(
            inputs=aig.inputs - set(inputs),
            node_map=aig.node_map if keep_outputs else frozenset(node_map),
            latch_map=aig.latch_map | _latch_map,
            latch2init=aig.latch2init | l2init
        )

    def unroll(self, horizon, *, init=True, omit_latches=True):
        # TODO:
        # - Check for name collisions.
        latches = self.latches
        aag0, l_map = self.cutlatches({l for l in latches})

        def _unroll():
            prev = aag0
            for t in range(1, horizon + 1):
                tmp = prev['i', {k: timed_name(k, t - 1) for k in aag0.inputs}]
                yield tmp['o', {k: timed_name(k, t) for k in aag0.outputs}]

        unrolled = reduce(seq_compose, _unroll())
        if init:
            source = {timed_name(n, 0): init for n, init in l_map.values()}
            unrolled = cmn.source(source) >> unrolled

        if omit_latches:
            latch_names = [timed_name(n, horizon) for n, _ in l_map.values()]
            unrolled = unrolled >> cmn.sink(latch_names)

        return unrolled

    def write(self, path):
        with open(path, 'w') as f:
            f.write(repr(self))

    def _modify_leafs(self, func):
        @fn.memoize
        def _mod(node):
            if isinstance(node, AndGate):
                left = _mod(node.left)
                right = _mod(node.right)
                if ConstFalse() in (left, right):
                    return ConstFalse()
                elif _is_const_true(left):
                    return right
                elif _is_const_true(right):
                    return left
                else:
                    return node._replace(left=left, right=right)

            elif isinstance(node, Inverter):
                child = _mod(node.input)
                if isinstance(child, Inverter):
                    return child.input
                else:
                    return node._replace(input=child)

            return func(node)

        node_map = ((name, _mod(cone)) for name, cone in self.node_map)
        latch_map = ((name, _mod(cone)) for name, cone in self.latch_map)
        return self.evolve(
            node_map=frozenset(node_map),
            latch_map=frozenset(latch_map)
        )


def par_compose(aig1, aig2, check_precondition=True):
    assert not aig1.latches & aig2.latches
    assert not aig1.outputs & aig2.outputs

    shared_inputs = aig1.inputs & aig2.inputs
    if shared_inputs:
        relabels1 = {n: cmn._fresh() for n in shared_inputs}
        relabels2 = {n: cmn._fresh() for n in shared_inputs}
        aig1, aig2 = aig1['i', relabels1], aig2['i', relabels2]

    circ = AIG(
        inputs=aig1.inputs | aig2.inputs,
        latch_map=aig1.latch_map | aig2.latch_map,
        latch2init=aig1.latch2init | aig2.latch2init,
        node_map=aig1.node_map | aig2.node_map,
        comments=aig1.comments + aig2.comments
    )

    if shared_inputs:
        for orig in shared_inputs:
            new1, new2 = relabels1[orig], relabels2[orig]
            circ <<= cmn.tee({orig: [new1, new2]})

    return circ


def seq_compose(circ1, circ2, *, input_kinds=(Input,)):
    interface = circ1.outputs & circ2.inputs
    assert not (circ1.outputs - interface) & circ2.outputs
    assert not circ1.latches & circ2.latches

    passthrough = {(k, v) for k, v in circ1.node_map if k not in interface}
    lookup = dict(circ1.node_map)

    def sub(node):
        if isinstance(node, input_kinds):
            return lookup.get(node.name, node)
        return node

    circ3 = circ2._modify_leafs(sub)
    return AIG(
        inputs=circ1.inputs | (circ2.inputs - interface),
        latch_map=circ1.latch_map | circ3.latch_map,
        latch2init=circ1.latch2init | circ2.latch2init,
        node_map=circ3.node_map | passthrough,
        comments=circ1.comments + circ2.comments
    )
